from duniterpy.documents import Block, BlockUID, Certification, Identity, Membership, \
    Revocation, Transaction, InputSource
from duniterpy.key import SigningKey
import time
import attr
import logging
import random
from ._user_identity import UserIdentity
from ._cert import Cert
from ._ms import MS
from ._ud import UD


@attr.s()
class BlockForge:
    """
    The block forge generates a blockchain with a PoW of 1
    """
    currency = attr.ib(validator=attr.validators.instance_of(str))
    key = attr.ib(validator=attr.validators.instance_of(SigningKey))
    pool = attr.ib(default=attr.Factory(list), validator=attr.validators.instance_of(list))
    blocks = attr.ib(default=attr.Factory(list), validator=attr.validators.instance_of(list))
    user_identities = attr.ib(default=attr.Factory(dict), validator=attr.validators.instance_of(dict))
    _ud = attr.ib(default=False, validator=attr.validators.instance_of(bool))
    _logger = attr.ib(default=attr.Factory(lambda: logging.getLogger('mirage')))

    @classmethod
    def start(cls, currency, salt, password, scrypt_params, loop):
        key = SigningKey(salt, password, scrypt_params)
        return cls(currency, key)

    def push(self, document):
        self.pool.append(document)

    def next_dividend(self):
        if self._ud:
            try:
                latest_dividend = next(reversed([b.ud for b in self.blocks if b.ud]))
                return int(latest_dividend * 1.1)
            except StopIteration:
                return 100
            finally:
                self._ud = False

    def previous_hash(self):
        try:
            latest_block = next(reversed(self.blocks))
        except StopIteration:
            return BlockUID.empty().sha_hash
        return latest_block.sha_hash

    def previous_issuer(self):
        try:
            latest_block = next(reversed(self.blocks))
        except StopIteration:
            return None
        return latest_block.issuer

    def members_count(self):
        return len([i for i in self.user_identities.values() if i.member])

    def identities(self):
        return [d for d in self.pool if type(d) is Identity]

    def revocations(self):
        return [r for r in self.pool if type(r) is Revocation]

    def joiners(self):
        return [d for d in self.pool if type(d) is Membership and d.membership_type == 'IN'
                and ((d.issuer in self.user_identities and not self.user_identities[d.issuer].member)
                     or d.issuer in [d.pubkey for d in self.pool if type(d) is Identity])]

    def actives(self):
        return [d for d in self.pool if type(d) is Membership and d.membership_type == 'IN'
                and d.issuer in self.user_identities and self.user_identities[d.issuer].member]

    def leavers(self):
        return [d for d in self.pool if type(d) is Membership and d.membership_type == 'OUT'
                and d.issuer in self.user_identities and self.user_identities[d.issuer].member]

    def excluded(self):
        return []

    def certifications(self):
        return [d for d in self.pool if type(d) is Certification]

    def transactions(self):
        return [d for d in self.pool if type(d) is Transaction]

    def parameters(self):
        if not self.blocks:
            return 0.1, 86400, 100000, 10800, 40, 2629800, 31557600, 1, 604800, 604800,\
                                                0.9, 15778800, 5, 12, 300, 25, 0.66

    def monetary_mass(self, number=None):
        mass = 0
        for b in self.blocks:
            if b.ud and (not number or b.number <= number):
                mass += b.ud * b.members_count
        return mass

    def set_member(self, pubkey, member):
        self._logger.info("Set {0} ({1}) as member : {2}".format(self.user_identities[pubkey].uid,
                                                                 pubkey[:5], member))
        self.user_identities[pubkey].member = member
        self.user_identities[pubkey].was_member = self.user_identities[pubkey].was_member or member

    def generate_dividend(self):
        self._logger.info("Generate dividend")
        self._ud = True

    def build_data(self):
        previous_hash = self.previous_hash()
        previous_issuer = self.previous_issuer()
        parameters = self.parameters()
        members_count = self.members_count()
        identities = self.identities()
        joiners = self.joiners()
        actives = self.actives()
        leavers = self.leavers()
        revocations = self.revocations()
        excluded = self.excluded()
        certifications = self.certifications()
        transactions = self.transactions()

        block = Block(10, self.currency, len(self.blocks), 1, int(time.time()),
                      int(time.time()), self.next_dividend(), 0, 
                      self.key.pubkey, 5, 5, 5, previous_hash, previous_issuer, 
                      parameters, members_count, identities, 
                      joiners, actives, leavers, 
                      revocations, excluded, certifications, 
                      transactions, "", 0, None)
        block.inner_hash = block.computed_inner_hash()
        return block

    def forge_block(self):
        block = self.build_data()
        block.sign([self.key])

        block.noonce = int(random.random()*10000000)
        self._logger.info("New block generated : {0}".format(block.blockUID))

        self.blocks.append(block)
        for identity in block.identities:
            self.user_identities[identity.pubkey] = UserIdentity(pubkey=identity.pubkey, uid=identity.uid,
                                                                 blockstamp=identity.timestamp,
                                                                 signature=identity.signatures[0])
            self._logger.info("New identity : {0}".format(self.user_identities[identity.pubkey]))

        if block.ud:
            for identity in self.user_identities.values():
                if identity.member:
                    identity.sources.append(InputSource(block.ud, block.unit_base, 'D', identity.pubkey, block.number))
                    identity.ud_generated.append(UD(amount=block.ud,
                                                    base=block.unit_base,
                                                    block_number=block.number,
                                                    time=block.mediantime))

        for certification in block.certifications:
            cert = Cert(from_identity=self.user_identities[certification.pubkey_from],
                        to_identity=self.user_identities[certification.pubkey_to],
                        signature=certification.signatures[0],
                        written_on=block.blockUID,
                        block=certification.timestamp.number,
                        mediantime=next(b.mediantime
                                        for b in self.blocks if b.number == certification.timestamp.number))
            self.user_identities[certification.pubkey_from].certs_sent.append(cert)
            self.user_identities[certification.pubkey_to].certs_received.append(cert)

        for membership in block.joiners + block.actives + block.leavers:
            self.user_identities[membership.issuer].memberships.append(MS(pubkey=membership.issuer,
                                                                          type=membership.membership_type,
                                                                          written_on=block.number,
                                                                          blockstamp=membership.membership_ts,
                                                                          timestamp=block.mediantime))

        for tx in block.transactions:
            receivers = [o.conditions.left.pubkey for o in tx.outputs
                         if o.conditions.left.pubkey != tx.issuers[0]]
            self.user_identities[tx.issuers[0]].tx_sent.append(tx)
            self.user_identities[receivers[0]].tx_received.append(tx)

        self.pool = []



import attr


@attr.s()
class UserIdentity:
    pubkey = attr.ib(repr=True)
    uid = attr.ib(repr=True)
    blockstamp = attr.ib(repr=True)
    signature = attr.ib()
    member = attr.ib(default=False)
    was_member = attr.ib(default=False)
    revoked = attr.ib(default=False)
    revoked_on = attr.ib(default=None)
    revocation_sig = attr.ib(default=None)
    sources = attr.ib(default=attr.Factory(list))
    certs_sent = attr.ib(default=attr.Factory(list))
    certs_received = attr.ib(default=attr.Factory(list))
    memberships = attr.ib(default=attr.Factory(list))
    tx_sent = attr.ib(default=attr.Factory(list))
    tx_received = attr.ib(default=attr.Factory(list))
    ud_generated = attr.ib(default=attr.Factory(list))




import attr


@attr.s()
class Cert:
    from_identity = attr.ib(repr=True)
    to_identity = attr.ib(repr=True)
    signature = attr.ib()
    written_on = attr.ib()
    block = attr.ib()
    mediantime = attr.ib()






from setuptools import setup, find_packages
import os
import re


setup(
    name='duniter-mirage',

    version="0.1.3",

    packages=find_packages(),

    author="inso",

    author_email="insomniak.fr@gmail.com",

    description="A python mock server of [duniter](https://github.com/duniter/duniter) API",

    long_description=open('README.md').read(),

    # Active la prise en compte du fichier MANIFEST.in
    include_package_data=True,
    url='https://github.com/Insoleet/mirage',

    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: French",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.5",
        "Topic :: Communications",
    ],
    dependency_links=("git+https://github.com/duniter/duniter-python-api.git@dev",)

)



import attr
import logging
from duniterpy.documents import Identity, Membership, BlockUID, Certification, \
    InputSource, OutputSource, Transaction, Unlock, SIGParameter
from duniterpy.documents.transaction import reduce_base
from duniterpy.grammars import output
from duniterpy.key import SigningKey, ScryptParams


@attr.s()
class User:
    """
    The user generates identities documents and sign them
    """
    SCRYPT_PARAMS = ScryptParams(2 ** 12, 16, 1)

    currency = attr.ib(validator=attr.validators.instance_of(str))
    uid = attr.ib(validator=attr.validators.instance_of(str))
    key = attr.ib(validator=attr.validators.instance_of(SigningKey))
    salt = attr.ib(validator=attr.validators.instance_of(str))
    password = attr.ib(validator=attr.validators.instance_of(str))
    blockstamp = attr.ib(validator=attr.validators.instance_of(BlockUID))
    _logger = attr.ib(default=attr.Factory(lambda: logging.getLogger('mirage')))

    @classmethod
    def create(cls, currency, uid, salt, password, blockstamp):
        return cls(currency, uid, SigningKey(salt, password, User.SCRYPT_PARAMS), salt, password, blockstamp)

    def identity(self):
        identity = Identity(10, self.currency, self.key.pubkey, self.uid, self.blockstamp, [])
        identity.sign([self.key])

        return identity

    def join(self, blockstamp):
        ms_doc = Membership(10, self.currency, self.key.pubkey, blockstamp,
                 'IN', self.uid, self.blockstamp, [])
        ms_doc.sign([self.key])
        return ms_doc

    def leave(self, blockstamp):
        ms_doc = Membership(10, self.currency, self.key.pubkey, blockstamp,
                 'OUT', self.uid, self.blockstamp, [])
        ms_doc.sign([self.key])
        return ms_doc

    def certify(self, other, blockstamp):
        cert = Certification(10, self.currency, self.key.pubkey, other.key.pubkey, blockstamp, [])
        cert.sign(self.identity(), [self.key])
        return cert

    def outputs_from_sources(self, amount, sources):
        # such a dirty algorithmm
        # shamelessly copy pasted from sakia
        def current_value(inputs, overhs):
            i = 0
            for s in inputs:
                i += s.amount * (10**s.base)
            for o in overhs:
                i -= o[0] * (10**o[1])
            return i

        amount, amount_base = reduce_base(amount, 0)
        current_base = max([src.base for src in sources])
        result_sources = []
        outputs = []
        overheads = []
        buf_sources = list(sources)
        while current_base >= 0:
            for s in [src for src in buf_sources if src.base == current_base]:
                test_sources = result_sources + [s]
                val = current_value(test_sources, overheads)
                # if we have to compute an overhead
                if current_value(test_sources, overheads) > amount * (10**amount_base):
                    overhead = current_value(test_sources, overheads) - int(amount) * (10**amount_base)
                    # we round the overhead in the current base
                    # exemple : 12 in base 1 -> 1*10^1
                    overhead = int(round(float(overhead) / (10**current_base)))
                    source_value = s.amount * (10**s.base)
                    out = int((source_value - (overhead * (10**current_base)))/(10**current_base))
                    if out * (10**current_base) <= amount * (10**amount_base):
                        result_sources.append(s)
                        buf_sources.remove(s)
                        overheads.append((overhead, current_base))
                        outputs.append((out, current_base))
                # else just add the output
                else:
                    result_sources.append(s)
                    buf_sources.remove(s)
                    outputs.append((s.amount, s.base))
                if current_value(result_sources, overheads) == amount * (10 ** amount_base):
                    return result_sources, outputs, overheads

            current_base -= 1
        raise ValueError("Not enough sources")

    def tx_outputs(self, receiver, outputs, overheads):
        total = []
        outputs_bases = set(o[1] for o in outputs)
        for base in outputs_bases:
            output_sum = 0
            for o in outputs:
                if o[1] == base:
                    output_sum += o[0]
            total.append(OutputSource(output_sum, base, output.Condition.token(output.SIG.token(receiver.key.pubkey))))

        overheads_bases = set(o[1] for o in overheads)
        for base in overheads_bases:
            overheads_sum = 0
            for o in overheads:
                if o[1] == base:
                    overheads_sum += o[0]
            total.append(OutputSource(overheads_sum, base, output.Condition.token(output.SIG.token(self.key.pubkey))))

        return total

    def send_money(self, amount, sources, receiver, blockstamp, message):

        result = self.outputs_from_sources(amount, sources)
        inputs = result[0]
        computed_outputs = result[1]
        overheads = result[2]

        unlocks = []
        for i, s in enumerate(sources):
            unlocks.append(Unlock(i, [SIGParameter(0)]))
        outputs = self.tx_outputs(receiver, computed_outputs, overheads)

        tx = Transaction(3, self.currency, blockstamp, 0,
                            [self.key.pubkey], inputs, unlocks,
                            outputs, message, None)
        tx.sign([self.key])
        return tx